"""
Module providing helpers to manipulate jobs
"""

import os
import shutil
import quixote
from quixote.build import GeneratorType
import panza.build
from panza.config import get_config
from textwrap import dedent
from subprocess import Popen, run, DEVNULL
import tempfile


class WorkspaceCreationError(Exception):
    """
    Exception class representing an error related to creating a workspace for a job
    """

    def __init__(self, cause: Exception):
        self.__cause__ = cause

    def __str__(self):
        return f"cannot create workspace: {self.__cause__.strerror}: {self.__cause__.filename}"


class EnvironmentBuildError(Exception):
    """
    Exception class representing an error related to the construction of an environment
    """

    def __init__(self, cause: Exception):
        self.__cause__ = cause

    def __str__(self):
        return f"cannot build environment: {str(self.__cause__)}"


class DataFetchingError(Exception):
    """
    Exception class representing an error related to the fetching of data needed by the job
    """

    def __init__(self, cause):
        self.__cause__ = cause

    def __str__(self):
        return f"cannot fetch data to analyze: {str(self.__cause__)}"


class _DockerDaemon:
    def __init__(self):
        self.dockerd = None
        self.root_dir = tempfile.TemporaryDirectory(prefix="panza_").name
        self.network_brige_name = os.path.basename(self.root_dir).replace('_', '-')

    def launch(self):
        script = dedent(
            f"""
            set -x

            ip link add name {self.network_brige_name} type bridge
            ip addr add 10.9.8.7/24 dev {self.network_brige_name}
            ip link set dev {self.network_brige_name} up
            """
        )
        run(["bash", "-c", script])
        command = dedent(
            f"""
            dockerd \
              --bridge={self.network_brige_name} \
              --data-root={self.root_dir}/data \
              --exec-root={self.root_dir}/exec \
              --host=unix://{self.root_dir}/docker.sock \
              --pidfile={self.root_dir}/docker.pid
              """
        )
        self.dockerd = Popen(["bash", "-c", command], stdout=DEVNULL, stderr=DEVNULL)

    def stop(self):
        if self.dockerd is not None:
            self.dockerd.kill()
            shutil.rmtree(self.root_dir)
            os.system(f"ip link set {self.network_brige_name} down")
            os.system(f"ip link delete {self.network_brige_name}")

    def socket_path(self) -> str:
        return f"{self.root_dir}/docker.sock"


class ExecutableJobHandle:
    """
    Class representing a job that is ready to be executed
    """

    def __init__(self, root_dir: str, job_info, blueprint: quixote.Blueprint):
        self.root_dir = root_dir
        self.job_info = job_info
        self.blueprint = blueprint

    def run(self):
        """
        Execute the job
        """
        image_name = f"test-{self.job_info['module_id']}-{self.job_info['activity_id']}"
        volumes = [(self.root_dir, "/moulinette")]

        dockerd = _DockerDaemon()
        if self.blueprint.allow_docker is True:
            dockerd.launch()
            volumes.append((dockerd.socket_path(), "/var/run/docker.sock"))

        volumes = map(lambda vol: f"-v {vol[0]}:{vol[1]}", volumes)
        os.system(f"docker run --rm -ti {' '.join(volumes)} {image_name} panza_executor.py")

        if self.blueprint.allow_docker is True:
            dockerd.stop()

    def cleanup(self):
        """
        Cleanup the job's working directory
        """
        shutil.rmtree(self.root_dir)


class DataAwaitingJobHandle:
    """
    Class representing a job whose environment is already built, and needs to fetch data to operate on
    """

    def __init__(self, root_dir: str, job_info, blueprint: quixote.Blueprint):
        self.root_dir = root_dir
        self.job_info = job_info
        self.blueprint = blueprint

    def fetch_data(self) -> ExecutableJobHandle:
        """
        Fetch the data on which this job will operate

        The fetcher functions specified in the blueprint will be used
        """
        delivery_path = os.path.join(self.root_dir, "rendu")
        resources_path = os.path.join(self.root_dir, "resources")
        with quixote.new_context(
                **get_config(),
                **self.job_info,
                delivery_path=delivery_path,
                resources_path=resources_path
        ):
            for fetcher in self.blueprint.fetchers:
                try:
                    fetcher()
                except Exception as e:
                    raise DataFetchingError(e)
        return ExecutableJobHandle(self.root_dir, self.job_info, self.blueprint)


class JobWorkspaceHandle:
    """
    Class representing a job workspace, that is, a job with its working directory and its resources
    """

    def __init__(self, root_dir: str, job_info, blueprint: quixote.Blueprint):
        self.root_dir = root_dir
        self.job_info = job_info
        self.blueprint = blueprint

    def build_job_environment(self) -> DataAwaitingJobHandle:
        """
        Build the environment specified by the blueprint for this job.

        The builder functions specified in the blueprint will be used to build a suitable Docker image
        """
        try:
            generator = panza.build.create_script_generator(GeneratorType.DOCKER)
            generator.generate_to_file(
                self.blueprint,
                os.path.join(self.root_dir, "Dockerfile"),
                extra_context={**get_config(), "resources_path": os.path.join(self.root_dir, "resources")}
            )
            os.system(
                f"docker build -t test-{self.job_info['module_id']}-{self.job_info['activity_id']} {self.root_dir}")
            os.remove(os.path.join(self.root_dir, "Dockerfile"))
        except Exception as e:
            raise EnvironmentBuildError(e)
        return DataAwaitingJobHandle(self.root_dir, self.job_info, self.blueprint)


class JobInfo:
    """
    Class representing a job information, that is, a directory containing the job's resources, and its metadata
    """

    def __init__(self, job_dir: str, job_info):
        self.job_dir = job_dir
        self.job_info = job_info

    def create_workspace(self) -> JobWorkspaceHandle:
        """
        Create a workspace in which the job can be executed.

        This will create a directory, copy the required files, and load the blueprint
        """
        job_name = f"{self.job_info['module_id']}-{self.job_info['activity_id']}-{self.job_info['group_id']}"
        root_dir = os.path.join(get_config()["root_dir"], job_name)

        try:
            shutil.copytree(self.job_dir, root_dir)
        except OSError as e:
            raise WorkspaceCreationError(e)

        blueprint = panza.BlueprintLoader.load_from_directory(root_dir)

        return JobWorkspaceHandle(root_dir, self.job_info, blueprint)
